#!/usr/bin/env python3
"""
build fio
=========

Build fio in fio.repository.path, using prefix fio.build.prefix.

Retargetable: True
------------------
"""
from pathlib import Path


def main(args, cijoe, step):
    """Install qemu"""

    commands = [
        "make clean",
        f"./configure --prefix={ cijoe.config.options['fio']['build']['prefix'] }",
        "make -j $(nproc)",
    ]
    for cmd in commands:
        err, _ = cijoe.run(
            cmd, cwd=Path(cijoe.config.options["fio"]["repository"]["path"])
        )
        if err:
            return err

    return err



#!/usr/bin/env python3
"""
    Library functions for performance requirements and normalization of metrics
"""
import copy
import dataclasses
import os
import re
from typing import List, Tuple

import yaml

from cijoe.core.errors import InvalidRangeError, UnknownUnitError

UNITS = {
    # general
    "": 1,  # no unit
    "B": 1,  # bytes
    "k": 1000,  # kilo
    "M": 1000**2,  # mega
    "G": 1000**3,  # giga
    # kibi
    "KiB": 1024**1,  # kibibytes
    "MiB": 1024**2,  # mibibytes
    "GiB": 1024**3,  # gibibytes
    "TiB": 1024**4,  # tibibytes
    # kilo
    "kB": 1000**1,  # kilobytes
    "MB": 1000**2,  # megabytes
    "GB": 1000**3,  # gigabytes
    "TB": 1000**4,  # gigabytes
    # time
    "nsec": 1 / 1000**3,  # nanoseconds
    "usec": 1 / 1000**2,  # microseconds
    "msec": 1 / 1000**1,  # milliseconds
    "sec": 1,  # seconds
    "min": 60,  # minutes
}


class Range:
    """
    Range implements parsing and validation of mathematical range notation,
    e.g. `[-5;100[` which translates to "must be >= -5 and < 100".
    """

    # pylint: disable=no-self-use
    # pylint: disable=too-few-public-methods

    _rng_re = re.compile(
        r"^(?P<elower>\[|\])\s*(?P<rstart>-inf|-?\d+(\.\d*)?)\s*;"  # [1.0;
        r"\s*(?P<rend>inf|-?\d+(\.\d*)?)\s*(?P<eupper>\[|\])"  # 1.0]
        rf"\s*(?P<unit>({'|'.join(UNITS)}))$"  # ms
    )

    def __init__(self, rng: str):
        match = self._rng_re.match(rng)
        if not match:
            raise InvalidRangeError(f'invalid syntax or unit for "{rng}"')

        rng_start = float(match["rstart"])
        rng_end = float(match["rend"])
        if rng_start > rng_end:
            raise InvalidRangeError(
                "expected lower bound <= upper bound, " f"{rng_start} <= {rng_end}"
            )

        # NOTE: _rng_re enforces that match["unit"] exists in UNITS.
        unit_val = UNITS[match["unit"]]

        self._rng_start = rng_start
        self._rng_end = rng_end
        self._elower = match["elower"]
        self._eupper = match["eupper"]
        self._unit = match["unit"]

        self._check_lower = self._make_check_lower(
            match["elower"], rng_start * unit_val
        )
        self._check_upper = self._make_check_upper(match["eupper"], rng_end * unit_val)

    def contains(self, val: float) -> bool:
        """Check whether n is contained in range.

        val must be given in the base unit of the measurement, e.g. seconds for
        time and bytes for storage.
        """
        return self._check_lower(val) and self._check_upper(val)

    def _make_check_lower(self, edge_lower: str, rng_start: float):
        if edge_lower == "[":
            return lambda n: n >= rng_start
        if edge_lower == "]":
            return lambda n: n > rng_start
        raise InvalidRangeError("invalid input _make_check_lower")

    def _make_check_upper(self, edge_upper: str, rng_end: float):
        if edge_upper == "[":
            return lambda n: n < rng_end
        if edge_upper == "]":
            return lambda n: n <= rng_end
        raise InvalidRangeError("invalid input _make_check_upper")

    def format_val(self, val: float) -> str:
        """Formats and returns val using the unit of the range.

        Example:
            range: "[250; 750]usec"
            val: 0.0005
            output: "500 usec"
        """

        val_conv = val / UNITS[self._unit]
        return f"{val_conv:.3f} {self._unit}"

    def __str__(self):
        return (
            f"{self._elower}{self._rng_start};"
            f"{self._rng_end}{self._eupper} {self._unit}"
        )


def to_base_unit(val: float, unit: str = "") -> float:
    """Converts val in the given unit to its base unit.
    Example:
        val: 100, unit: 'KiB'
        output: 102400 (bytes)

        val: 500, unit: 'msec'
        output: 0.5 (seconds)
    """
    unit_scalar = UNITS.get(unit, None)
    if not unit_scalar:
        raise UnknownUnitError(f"Unit '{unit}' is not supported")

    return val * unit_scalar



"""

"""
import errno
import logging as log
import os
import time
import traceback
from pathlib import Path

import yaml

from cijoe.core import transport
from cijoe.core.misc import ENCODING, sanitize_ident
from cijoe.core.resources import Config


def default_output_path():
    """Returns a default output-path"""

    return os.path.join(os.getcwd(), "cijoe-output")


class CommandState(object):
    def __init__(
        self, cmd, cwd, err, begin, end, output_dpath, output_fpath, state_fpath
    ):
        self.cmd = cmd
        self.cwd = cwd
        self.err = err
        self.begin = begin
        self.end = end
        self.elapsed = end - begin
        self.output_dpath = Path(output_dpath)
        self.output_fpath = Path(output_fpath)
        self.state_fpath = Path(state_fpath)

    def output(self):
        """Returns the content of 'output_fpath'"""

        with self.output_fpath.open() as ofd:
            return ofd.read()

    def to_file(self):
        """Dump the command state to file"""

        with self.state_fpath.open("a", encoding=ENCODING) as state_file:
            state = {
                "cmd": self.cmd,
                "cwd": str(self.cwd),
                "err": self.err,
                "begin": self.begin,
                "end": self.end,
                "elapsed": self.elapsed,
                "output_dpath": str(self.output_dpath),
                "output_fpath": str(self.output_fpath),
            }
            yaml.dump(state, state_file)


class Cijoe(object):
    """CIJOE providing retargetable command-line expressions and data-transfers"""

    def __init__(self, config: Config, output_path: Path):
        """Create a cijoe encapsulation defined by the given config_fpath"""

        self.config = config

        self.run_count = 0
        self.output_path = output_path if output_path else default_output_path()
        self.output_ident = "artifacts"

        os.makedirs(os.path.join(self.output_path, self.output_ident), exist_ok=True)

        self.transport_local = transport.Local(self.config, self.output_path)
        self.transport = self.transport_local

        ssh = self.config.options.get("cijoe", {}).get("transport", {}).get("ssh", None)
        if ssh:
            self.transport = transport.SSH(self.config, self.output_path)

    def set_output_ident(self, output_ident: str):
        """
        This sets the output-identifier which is used in order to provide a subfolder
        for artifacts, command-output etc. Additionally, then it reset the command
        run-count
        """

        output_ident = sanitize_ident(output_ident)

        self.run_count = 0
        self.output_ident = output_ident
        self.transport.output_ident = output_ident

    def _run(self, cmd, cwd, env, transport):
        self.run_count += 1
        cmd_output_dpath = os.path.join(self.output_path, self.output_ident)
        cmd_output_fpath = os.path.join(
            cmd_output_dpath, f"cmd_{self.run_count:02}.output"
        )
        cmd_state_fpath = os.path.join(
            cmd_output_dpath, f"cmd_{self.run_count:02}.state"
        )
        os.makedirs(cmd_output_dpath, exist_ok=True)

        with open(cmd_output_fpath, "a", encoding=ENCODING) as logfile:
            begin = time.time()
            err = transport.run(cmd, cwd, env, logfile)
            state = CommandState(
                cmd=cmd,
                cwd=cwd,
                err=err,
                begin=begin,
                end=time.time(),
                output_dpath=cmd_output_dpath,
                output_fpath=cmd_output_fpath,
                state_fpath=cmd_state_fpath,
            )
            state.to_file()

        return err, state

    def run(self, cmd, cwd=None, env={}):
        """
        Execute the given shell command/expression via 'config.transport'

        Commands executed using this will write stdout and stderr to file. The location
        of the logfile is fixed to: "output_path/output_ident/cmd.log", such that the
        location is a subfolder of the output_path. Unless somebody wants to break the
        convention and call set_output_ident("../..")
        """

        return self._run(cmd, cwd, env, self.transport)

    def run_local(self, cmd, cwd=None, env={}):
        """
        Execute the given shell command/expression via local transport

        Commands executed using this will write stdout and stderr to file. The location
        of the logfile is fixed to: "output_path/output_ident/cmd.log", such that the
        location is a subfolder of the output_path. Unless somebody wants to break the
        convention and call set_output_ident("../..")
        """

        return self._run(cmd, cwd, env, self.transport_local)

    def put(self, src, dst):
        """Transfer 'src' on 'dev_box' to 'dst' on **test_target**"""

        os.makedirs(os.path.join(self.output_path, self.output_ident), exist_ok=True)

        try:
            return self.transport.put(src, dst)
        except Exception as exc:
            log.error(f"err({exc})")
            log.debug(f"src({src}), dst({dst})")
            log.debug(
                f"run():{type(exc).__name__}, {__file__}:{exc.__traceback__.tb_lineno}"
            )

        return False

    def get(self, src, dst):
        """Transfer 'src' on 'test_target' to 'dst' on **dev_box**"""

        os.makedirs(os.path.join(self.output_path, self.output_ident), exist_ok=True)

        try:
            return self.transport.get(src, dst)
        except Exception as exc:
            log.error(f"err({exc})")
            log.debug(f"src({src}), dst({dst})")
            log.debug(
                f"run():{type(exc).__name__}, {__file__}:{exc.__traceback__.tb_lineno}"
            )

        return False



import argparse
import errno
import jinja2
import logging as log
import os
import shutil
import sys
import tempfile
import time
from pathlib import Path

import cijoe.core
from cijoe.core.command import Cijoe, default_output_path
from cijoe.core.monitor import WorkflowMonitor
from cijoe.core.resources import (
    Config,
    Workflow,
    dict_from_yamlfile,
    dict_substitute,
    get_resources,
)

DEFAULT_CONFIG_FILENAME = "cijoe-config.toml"
DEFAULT_WORKFLOW_FILENAME = "cijoe-workflow.yaml"
DEFAULT_SCRIPT_FILENAME = "cijoe-script.py"
SEARCH_PATHS = [
    Path.cwd(),
    Path.cwd() / ".cijoe",
    Path.home() / ".config" / "cijoe",
    Path.home() / ".cijoe",
]


def search_for_file(path: Path):
    """
    Search for a file named path.name. In case the given 'path' does not
    have a directory path, then SEARCH_PATH is used.

    Returns a resolved Path object on success and None on error.
    """

    # Has a directory part; resolve it and check that it exists
    if path.is_absolute() or len(path.parts) != 1:
        path = path.resolve()
        path = path if path.exists() else None

        return path

    # Only a filename; look in search paths
    for spath in SEARCH_PATHS:
        candidate = (spath / path).resolve()
        if candidate.exists():
            return candidate

    return None


def log_errors(errors):
    for error in errors:
        log.error(error)


def cli_integrity_check(args):
    """Lint a workflow"""

    log.info("cli: lint")
    log.info(f"workflow: '{args.workflow}'")
    log.info(f"config: '{args.config}'")

    workflow_dict = dict_from_yamlfile(args.workflow.resolve())
    errors = Workflow.dict_normalize(workflow_dict)  # Normalize it
    errors += Workflow.dict_lint(workflow_dict)  # Check the yaml-file

    config = Config.from_path(args.config)
    if not config:
        log.error(f"failed: Config.from_path({args.config})")
        return errno.EINVAL

    errors += dict_substitute(workflow_dict, config.options)

    if errors:
        log_errors(errors)
        log.error("failed: 'see errors above'; Failed")
        return errno.EINVAL

    return 0


def cli_resources(args):
    """List the reference configuration files provided with cijoe packages"""

    log.info("cli: resources")
    resources = get_resources()

    print("Resources collected by the CIJOE collector are listed below.")
    for category, category_resources in sorted(resources.items()):
        print(f"{category}:" + ("" if category_resources.items() else " ~"))

        for ident, path in sorted(category_resources.items()):
            print(f"  - ident: {ident}")
            print(f"    path: {path}")

    return 0


def cli_archive(args):
    """Move 'output' directory into archive"""

    if args.output.exists():
        state_path = args.output / "workflow.state"
        tag = ""
        if state_path.exists():
            state = dict_from_yamlfile(state_path)
            tag = "" if state.get("tag", None) is None else "-".join(state["tag"]) + "-"
        t = str(time.strftime("%Y-%m-%d_%H:%M:%S"))
        archive = args.output.with_name("cijoe-archive") / f"{tag}{t}"
        os.makedirs(archive)
        log.info(f"moving existing output-directory({args.output}) to '{archive}'")
        os.rename(args.output, archive)


def cli_produce_report(args):
    """Produce workflow-report"""

    config_path = args.output / "config.orig"
    state_path = args.output / "workflow.state"

    if not state_path.exists():
        log.error("no workflow.state, nothing to produce a report for")
        return errno.EINVAL

    if not config_path.exists():
        log.error("missing config")
        return errno.EINVAL

    # Check config/substitutions
    config = Config.from_path(config_path)
    if not config:
        log.error(f"failed: Config.from_path({config_path})")
        return errno.EINVAL

    cijoe = Cijoe(config, args.output)

    resources = get_resources()

    reporter = resources["scripts"]["core.reporter"]
    if reporter.func is None:
        reporter.load()

    return reporter.func(
        args,
        cijoe,
        {
            "name": "report",
            "uses": "core.reporter",
            "with": {"report_open": args.skip_report},
        },
    )


def cli_example(args):
    """Create example config.toml and workflow.yaml"""

    log.info("cli: examples")
    err = 0

    resources = get_resources()
    print(resources["configs"].keys())

    resource = resources["configs"].get(f"{args.example}.default-config", None)
    if resource is None:
        log.error(
            f"'default-config{Config.SUFFIX}' from '{args.example}'; not available"
        )
        return errno.EINVAL
    src_config = resource.path

    resource = resources["workflows"].get(f"{args.example}.example-workflow", None)
    if resource is None:
        log.error(
            f"'example-workflow{Workflow.SUFFIX}' from '{args.example}'; not available"
        )
        return errno.EINVAL
    src_workflow = resource.path

    resource = resources["scripts"].get("core.example", None)
    if resource is None:
        log.error(
            f"example.py' from '{args.example}'; not available"
        )
        return errno.EINVAL
    src_script = resource.path

    dst_config = Path.cwd().joinpath(DEFAULT_CONFIG_FILENAME)
    dst_workflow = Path.cwd().joinpath(DEFAULT_WORKFLOW_FILENAME)
    dst_script = Path.cwd().joinpath(DEFAULT_SCRIPT_FILENAME)

    log.info(f"config: {dst_config}")
    log.info(f"workflow: {dst_workflow}")
    log.info(f"script: {dst_script}")

    if not src_config.exists():
        log.error(
            f"'default-config{Config.SUFFIX}' from '{args.example}'; not available"
        )
        return errno.EINVAL
    if not src_workflow.exists():
        log.error(
            f"example-workflow{Workflow.SUFFIX}' from '{args.example}'; not available"
        )
        return errno.EINVAL
    if not src_script.exists():
        log.error(
            f"example.py' from '{args.example}'; not available"
        )
        return errno.EINVAL

    if dst_config.exists():
        err = errno.EEXIST
        log.error(f"skipping config({dst_config}); already exists")
    else:
        shutil.copyfile(src_config, dst_config)

    if dst_workflow.exists():
        err = errno.EEXIST
        log.error(f"skipping workflow({dst_workflow}); already exists")
    else:
        shutil.copyfile(src_workflow, dst_workflow)

    if dst_script.exists():
        err = errno.EEXIST
        log.error(f"skipping script({dst_script}); already exists")
    else:
        shutil.copyfile(src_script, dst_script)

    return err


def cli_version(args):
    """Print version and exit"""

    print(f"cijoe {cijoe.core.__version__}")

    return 0


def cli_monitor(args):
    """Start workflow monitor"""

    monitor = WorkflowMonitor()
    monitor.start()
    monitor.join()

    return 0


def cli_workflow(args):
    """Process workflow"""

    log.info("cli: run")
    log.info(f"workflow: {args.workflow}")
    log.info(f"config: {args.config}")
    log.info(f"output: {args.output}")

    cli_archive(args)

    state_path = args.output / "workflow.state"
    if state_path.exists():
        log.error(f"aborting; output({args.output}) directory already exists")
        return errno.EPERM

    config = Config(args.config)
    errors = config.load()
    if errors:
        log_errors(errors)
        log.error("failed: Config(args.config).load()")
        return errno.EINVAL

    workflow = Workflow(args.workflow)

    errors = workflow.load(config)
    if errors:
        log_errors(errors)
        log.error("workflow.load(): see errors above or run 'cijoe -i'")
        return errno.EINVAL

    workflow.state["tag"] = args.tag
    step_names = [step["name"] for step in workflow.state["steps"]]
    for step_name in args.step:
        if step_name in step_names:
            continue

        log.error(f"step({step_name}) not in workflow")
        return errno.EINVAL

    os.makedirs(args.output)
    shutil.copyfile(args.config, args.output / "config.orig")
    shutil.copyfile(args.workflow, args.output / "workflow.orig")
    resources = get_resources()

    # pre-load scripts and augment state with step-descriptions.
    # TODO: for some reason when mod.__doc__ is None, then the docstring from a previous
    # mod trickles in. This should be fixed...
    for step in workflow.state["steps"]:
        script_ident = step["uses"]
        resources["scripts"][script_ident].load()

        step["description"] = "Undocumented"
        if resources["scripts"][script_ident].mod.__doc__:
            step["description"] = str(resources["scripts"][script_ident].mod.__doc__)

    workflow.state["status"]["started"] = time.time()

    workflow.state_dump(args.output / Workflow.STATE_FILENAME)

    fail_fast = False

    cijoe = Cijoe(config, args.output)
    for step in workflow.state["steps"]:
        log.info(f"step({step['name']}) - begin")

        begin = time.time()

        cijoe.set_output_ident(step["id"])
        os.makedirs(os.path.join(cijoe.output_path, step["id"]), exist_ok=True)

        if args.step and step["name"] not in args.step:
            step["status"]["skipped"] = 1
        else:
            script_ident = step["uses"]

            try:
                err = resources["scripts"][script_ident].func(args, cijoe, step)
                if err:
                    log.error(f"script({script_ident}) : err({err})")
                step["status"]["failed" if err else "passed"] = 1
            except KeyboardInterrupt:
                log.exception(f"script({script_ident}) : failed")
                step["status"]["failed"] = 1
            except Exception:
                log.exception(f"script({script_ident}) : failed")
                step["status"]["failed"] = 1

        for key in ["failed", "passed", "skipped"]:
            workflow.state["status"][key] += step["status"][key]

        step["status"]["elapsed"] = time.time() - begin
        workflow.state["status"]["elapsed"] += step["status"]["elapsed"]
        workflow.state_dump(args.output / Workflow.STATE_FILENAME)

        for text, status in step["status"].items():
            if text != "elapsed" and status:
                log.info(f"step({step['name']}) : {text}")

        if step["status"]["failed"] and fail_fast:
            log.error(f"exiting, fail_fast({fail_fast})")
            break

    err = errno.EIO if workflow.state["status"]["failed"] else 0
    if err:
        log.error("one or more steps failed")

    return err


def create_adhoc_workflow(args, paths):
    resources = get_resources()
    paths = map(Path, paths)

    template_path = resources["templates"]["core.example-tmp-workflow.yaml"].path
    jinja_env = jinja2.Environment(
        autoescape=True, loader=jinja2.FileSystemLoader(template_path.parent)
    )
    template = jinja_env.get_template(template_path.name)

    with tempfile.NamedTemporaryFile() as workflow:
        setattr(args, "workflow", Path(workflow.name))
        setattr(args, "step", [])

        content = template.render(paths=paths)
        workflow.write(bytes(content, "utf-8"))
        workflow.seek(0)

        sys.exit(main(args))


def cli_interface(path):
    path = Path(path)
    args = parse_args()
    create_adhoc_workflow(args, [path])


def parse_args():
    """Parse command-line interface."""

    parser = argparse.ArgumentParser(
        prog=Path(sys.argv[0]).stem,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    workflow_group = parser.add_argument_group(
        "workflow", "Run workflow at '-w', using config at '-c', and output at '-o'"
    )

    workflow_group.add_argument(
        "step", nargs="*", help="Given a workflow; one or more workflow steps to run. Else; one or more cijoe Python scripts to run."
    )

    workflow_group.add_argument(
        "--config",
        "-c",
        type=Path,
        default=Path(os.environ.get("CIJOE_DEFAULT_CONFIG", DEFAULT_CONFIG_FILENAME)),
        help="Path to the Configuration file.",
    )
    workflow_group.add_argument(
        "--workflow",
        "-w",
        type=Path,
        default=Path(
            os.environ.get("CIJOE_DEFAULT_WORKFLOW", DEFAULT_WORKFLOW_FILENAME)
        ),
        help="Path to workflow file.",
    )
    workflow_group.add_argument(
        "--output",
        "-o",
        type=Path,
        default=default_output_path(),
        help="Path to output directory.",
    )
    workflow_group.add_argument(
        "--log-level",
        "-l",
        action="append_const",
        const=1,
        help="Increase log-level. Provide '-l' for info and '-ll' for debug.",
    )
    workflow_group.add_argument(
        "--no-report",
        "-n",
        action="store_true",
        help="Skip the producing, and opening, a report at the end of the workflow-run",
    )
    workflow_group.add_argument(
        "--skip-report",
        "-s",
        action="store_false",
        help="Skip the report opening at the end of the workflow-run",
    )
    workflow_group.add_argument(
        "--tag",
        "-t",
        type=str,
        action="append",
        help="Tags to identify a workflow-run."
        " This will be prefixed while storing in archive",
    )

    utils_group = parser.add_argument_group(
        "utilities", "Workflow, and workflow-related utilities"
    )
    utils_group.add_argument(
        "--archive",
        "-a",
        action="store_true",
        help="Move the output at '-o / --output' to 'cijoe-archive/YYYY-MM-DD_HH:MM:SS",
    )
    utils_group.add_argument(
        "--produce-report",
        "-p",
        action="append_const",
        const=1,
        help="Produce report, and open it, for output at '-o / --output' and exit.",
    )
    utils_group.add_argument(
        "--monitor",
        "-m",
        action="store_true",
        help="Monitor workflow-output in the current-workdir-directory (cwd).",
    )
    utils_group.add_argument(
        "--integrity-check",
        "-i",
        action="store_true",
        help="Check integrity of workflow at '-w / --workflow' and exit.",
    )
    utils_group.add_argument(
        "--resources",
        "-r",
        action="store_true",
        help="List collected resources and exit.",
    )
    utils_group.add_argument(
        "--example",
        "-e",
        action="store",
        const="core",
        type=str,
        nargs="?",
        default=None,
        help="Create 'default-config.toml', 'example-workflow.yaml' and 'cijoe-script.py' and exit.",
    )
    utils_group.add_argument(
        "--version",
        "-v",
        action="store_true",
        help="Print the version number of 'cijoe' and exit.",
    )

    return parser.parse_args()


def main(args=None):
    """Main entry point for the CLI"""

    if args is None:
        args = parse_args()
        if args.step and all(step.endswith(".py") for step in args.step):
            create_adhoc_workflow(args, args.step)

    levels = [log.ERROR, log.INFO, log.DEBUG]
    log.basicConfig(
        format="%(levelname)s:%(module)s:%(funcName)s(): %(message)s",
        level=levels[
            min(sum(args.log_level), len(levels) - 1) if args.log_level else 0
        ],
    )

    if args.resources:
        return cli_resources(args)

    if args.example:
        return cli_example(args)

    if args.version:
        return cli_version(args)

    if args.produce_report:
        return cli_produce_report(args)

    if args.archive:
        return cli_archive(args)

    if args.monitor:
        return cli_monitor(args)

    for filearg in ["config", "workflow"]:
        argv = getattr(args, filearg)
        path = search_for_file(argv)
        if path is None:
            log.error(f"{filearg}({argv}) does not exist; exiting")
            return errno.EINVAL

        setattr(args, filearg, path)

    if args.integrity_check:
        return cli_integrity_check(args)

    # At this point all Path objects are resolved, except for output, do that
    # that here.
    args.output = args.output.resolve()

    err = cli_workflow(args)

    if not args.no_report:
        report_err = cli_produce_report(args)
        err = err if err else report_err

    return err



import pytest

import cijoe.linux.null_blk as null_blk


def skip_when_config_has_no_remote(cijoe):
    """Skip testing when configuration is module not enabled"""

    transport = cijoe.config.options.get("cijoe", {}).get("transport", None)
    if not transport:
        pytest.skip(reason="skipping as there is no remote transport defined")


def test_insert(cijoe):
    """Test the creation of null_block via module-load"""

    skip_when_config_has_no_remote(cijoe)

    config = cijoe.config.options.get("null_blk", None)
    assert config, "Invalid environment configuration"

    nr_devices = int(config.get("nr_devices"))
    assert nr_devices, "!nr_devices, only module-load instances are supported"

    err, _ = null_blk.insert(cijoe)
    assert not err, "Failed inserting kernel module"

    err, _ = cijoe.run("lsblk")
    assert not err, "Failed listing block devices"

    for n in range(nr_devices):
        err, _ = cijoe.run(f"file /dev/nullb{n}")
        assert not err


def test_remove(cijoe):

    skip_when_config_has_no_remote(cijoe)

    err, _ = null_blk.remove(cijoe)
    assert not err, "Failed removing kernel module"



"""
Report generator
================

Generates a HTML report in the workflow output directory.

Retargtable: false
------------------

The report-generator works on the files generated by cijoe on the host which is
executing cijoe. Thus, no need to make this re-targetable.

Step arguments
--------------

# Whether or not the generated report should be opened (in a browser)
report_open: true|false
"""
import logging as log
import webbrowser
from datetime import datetime

import jinja2
import yaml

from cijoe.core.processing import process_workflow_output
from cijoe.core.resources import get_resources


def to_yaml(value):
    return yaml.dump(value)


def elapsed_txt(value):
    minutes, seconds = divmod(float(value), 60.0)
    hours, minutes = divmod(minutes, 60.0)

    txt = []
    if hours:
        txt.append(f"{hours:.0f} hour")
    if minutes:
        txt.append(f"{minutes:.0f} min")
    txt.append(f"{seconds:0.2f} sec")

    return " ".join(txt)


def timestamp_to_txt(value):
    return datetime.fromtimestamp(float(value)).strftime("%d-%m-%Y, %H:%M:%S")


def main(args, cijoe, step):
    """Produce a HTML report of the 'workflow.state' file in 'args.output'"""

    report_open = step.get("with", {"report_open"})["report_open"]

    resources = get_resources()

    template_path = resources["templates"]["core.report-workflow"].path
    report_path = args.output / "report.html"

    log.info(f"template: {template_path}")
    log.info(f"report: {report_path}")

    workflow_state = process_workflow_output(args, cijoe)

    jinja_env = jinja2.Environment(
        autoescape=True, loader=jinja2.FileSystemLoader(template_path.parent)
    )
    jinja_env.filters["to_yaml"] = to_yaml
    jinja_env.filters["elapsed_txt"] = elapsed_txt
    jinja_env.filters["timestamp_to_txt"] = timestamp_to_txt
    template = jinja_env.get_template(template_path.name)

    with (report_path).open("w") as report:
        report.write(template.render(workflow_state))

    if report_open:
        webbrowser.open("file://%s" % report_path.resolve())

    return 0



import errno
from pathlib import Path

import requests

ENCODING = "UTF-8"


def sanitize_ident(ident: str):
    """Naive ident sanitizer"""

    for illegal in r'/\|!?[]<>.,:;"*':
        ident = ident.replace(illegal, "_")

    return ident


def download(url: str, path: Path):
    """Downloads a file over http(s), returns (err, path)."""

    path = Path(path).resolve()
    if path.is_dir():
        path = path / url.split("/")[-1]
    if not (path.parent.is_dir() and path.parent.exists()):
        return errno.EINVAL, path

    with requests.get(url, stream=True) as request:
        request.raise_for_status()
        with path.open("wb") as local:
            for chunk in request.iter_content(chunk_size=8192):
                local.write(chunk)

    return 0, path



"""
cmdrunner
=========

Executes a list of commands in the given order. Note that multi-line commands are not
support, each line or list of strings are treated as individual commands.

Retargetable: True
------------------
"""
import errno


def main(args, cijoe, step):
    """Run commands one at a time via cijoe.run()"""

    err = 0
    if not ("with" in step and "commands" in step["with"]):
        return errno.EINVAL

    for cmd in step["with"]["commands"]:
        err, state = cijoe.run(cmd)
        if err:
            break

    return err



#!/usr/bin/env python3
"""
Start a qemu guest
==================


Retargetable: false
-------------------
"""
import errno
import logging as log

from cijoe.qemu.wrapper import Guest


def main(args, cijoe, step):
    """Start a qemu guest"""

    guest = Guest(cijoe, cijoe.config)

    err = guest.start()
    if err:
        log.error(f"guest.start() : err({err})")
        return err

    started = guest.is_up()
    if not started:
        log.error("guest.is_up() : False")
        return errno.EAGAIN

    return 0



import cijoe.linux.null_blk as null_blk
import pytest

import cijoe.fio.wrapper as fio


def skip_when_config_has_no_remote(cijoe):
    """Skip testing when configuration is module not enabled"""

    transport = cijoe.config.options.get("cijoe", {}).get("transport", None)
    if not transport:
        pytest.skip(reason="skipping as there is no remote transport defined")


def test_run(cijoe):

    skip_when_config_has_no_remote(cijoe)

    err, _ = null_blk.insert(cijoe)
    assert not err

    err, _ = fio.run(
        cijoe,
        [
            "--filename",
            "/dev/nullb0",
            "--bs",
            "4k",
            "--rw",
            "randread",
            "--size",
            "1G",
            "--name",
            "foo42",
        ],
    )
    assert not err



#!/usr/bin/env python
"""
Monitor a directory for creation and modification of cijoe command-output files
and dump their content to stdout. It is intentional that this should be
executed standalone like so::

    cijoe -m

That is, a command/process to start and have running in a terminal across
multiple invocations of cijoe. Thereby separating the output from the
instrumentation.

.. note::
    This is a quick-n-dirty prototype. Not much error-handling is done here, so
    expect crashing in case of non-printable characters and potential
    un-expected behavior.

.. note::
    In a perfect world, then this would not need to be implemented, instead
    tail / xtail could be used. However, those excellent tools are just not
    readily available everywhere.
"""
import logging as log
import queue
import re
import threading
import time
from pathlib import Path

from watchdog.events import FileSystemEventHandler
from watchdog.observers import Observer


def worker(monitor):
    """Worker dumping the contents of cijoe command-output files"""

    bytes_read = {}

    while True:
        task, path = monitor.workq.get()

        ident = str(path)  # Output identifier
        if ident not in bytes_read:
            bytes_read[ident] = 0

        if task == "print":
            with path.open(
                "rb"
            ) as cmdlog:  # Read, update stats, inform, dump, and update stats
                cmdlog.seek(bytes_read[ident])
                buf = cmdlog.read()
                nbytes = len(buf)
                log.info(
                    f"# cijoe.monitor: output({path})@{bytes_read[ident]}+{nbytes}"
                )
                try:
                    print(buf.decode("UTF-8"), end="")
                except UnicodeError:
                    pass
                monitor.dumped_at = time.time()
                bytes_read[ident] += nbytes

        monitor.workq.task_done()


class CommandOutputHandler(FileSystemEventHandler):
    """
    Handler filtering out creation and modification events for cijoe
    command-output files, forwarding these events to the worker to dump the
    file-content.
    """

    REGEX_CMDLOG_FILENAME = r"cmd_\d+\.output"

    def __init__(self, workq):
        self.workq = workq

    def skipevent(self, event):
        path = Path(event.src_path).resolve()
        if not re.match(CommandOutputHandler.REGEX_CMDLOG_FILENAME, str(path.name)):
            return True
        if event.is_directory:
            return True

        return False

    def on_created(self, event):
        """When a file is created, then reset the meta-data"""

        if self.skipevent(event):
            return

        self.workq.put(("reset", Path(event.src_path)))

    def on_modified(self, event):
        """When a file is modified, then dump the additions to stdout"""

        if self.skipevent(event):
            return

        self.workq.put(("print", Path(event.src_path)))


class WorkflowMonitor(object):
    """
    Monitors a directory path, recursively, for the creation and modification
    of cijoe commmand-output files for the purpose of dumping the file-content
    to stdout.
    """

    def __init__(self, path: Path = None):
        if path is None:
            path = Path.cwd()

        self.path = path
        self.observer = Observer()
        self.dumped_at = None
        self.dumped_notify = 3.0

        self.workq = queue.Queue()
        self.workt = threading.Thread(
            kwargs={"monitor": self}, target=worker, daemon=True
        )

    def start(self):
        log.info(f"# cijoe.monitor: Recursively observing path({self.path})")

        self.workt.start()

        self.observer.schedule(
            CommandOutputHandler(self.workq), str(self.path), recursive=True
        )
        self.observer.start()

    def stop(self):
        self.observer.stop()
        self.observer.join()

        self.workt.stop()

    def join(self):
        try:
            while True:
                time.sleep(1)

                if self.dumped_at is None:
                    log.info(
                        "# cijoe.monitor: no command-output yet; start your workflow?"
                    )
                    continue

                when = self.dumped_at
                seconds = time.time() - when
                if seconds >= self.dumped_notify:
                    log.info(
                        f"# cijoe.monitor: no output for seconds({seconds:.2f}) .."
                    )
                    if seconds >= self.dumped_notify * 2:
                        log.info(
                            "# cijoe.monitor: .. possibly just a silent command .."
                        )
                    if seconds >= self.dumped_notify * 3:
                        log.info("# cijoe.monitor: .. or, workflow is finished ..")
                    if seconds >= self.dumped_notify * 4:
                        log.info(
                            "# cijoe.monitor: .. run 'cijoe -p' to see the progress"
                        )

        except KeyboardInterrupt:
            pass



"""
    Helper functions for processing the output produced by a workflow
"""
import json
from pathlib import Path

from cijoe.core.misc import ENCODING, sanitize_ident
from cijoe.core.resources import dict_from_yamlfile


def runlog_from_path(path: Path):
    """Produce a dict of command-dicts with paths to .output and .state files"""

    run = {}

    if not (path.is_dir() and path.exists()):
        return run

    for cmd_path in sorted(path.glob("cmd_*.*")):
        stem = cmd_path.stem
        suffix = cmd_path.suffix[1:]
        if suffix not in ["output", "state"]:
            continue

        if stem not in run:
            run[stem] = {
                "output_path": None,
                "output": "",
                "state": {},
                "state_path": None,
            }

        run[stem][f"{suffix}_path"] = cmd_path
        if suffix == "output":
            with run[stem][f"{suffix}_path"].open(
                encoding=ENCODING, errors="replace"
            ) as content:
                run[stem][f"{suffix}"] = content.read()
        elif suffix == "state":
            run[stem][f"{suffix}"] = dict_from_yamlfile(run[stem][f"{suffix}_path"])

    return run


def longrepr_to_string(longrepr):
    """Extract pytest crash/traceback/longrepr info from pytest structure"""

    lines = []

    lines.append("# crashinfo")
    reprcrash = longrepr.get("reprcrash", {})
    for key, value in reprcrash.items():
        lines.append(f"{key}: {value}")

    entries = longrepr.get("reprtraceback", {"reprentries": []}).get("reprentries", [])
    for entry in entries:
        if entry is None:
            continue

        data = entry.get("data")
        if data is None:
            continue

        reprfuncargs = data.get("reprfuncargs")
        if reprfuncargs is None:
            continue

        reprargs = reprfuncargs.get("args")
        if reprargs is None:
            continue

        lines.append("")
        lines.append("# test-args")

        for argline in reprargs:
            lines.append(":".join(argline))

        lines.append("")
        lines.append("# test-output-lines")
        for dataline in entry.get("data", {"lines": []}).get("lines", []):
            lines.append(dataline)

    return "\n".join(lines)


def testreport_from_file(path: Path):
    """Parse the given 'pytest-reportlog' output into a restreport dict"""

    results = {
        "status": {"failed": 0, "passed": 0, "skipped": 0, "total": 0},
        "tests": {},
    }

    logpath = path / "testreport.log"
    if not logpath.exists():
        return {}

    with logpath.open() as logfile:
        for count, line in enumerate(logfile.readlines()):
            result = json.loads(line)
            if result["$report_type"] != "TestReport":
                continue

            nodeid = result["nodeid"]
            if nodeid not in results["tests"]:
                try:
                    comp = nodeid.split("::")
                    group_left = comp[0]
                    group_right = "".join(comp[1:])
                except Exception:
                    group_left, group_right = (nodeid, nodeid)

                results["tests"][nodeid] = {
                    "group_left": group_left,
                    "group_right": group_right,
                    "count": count,
                    "nodeid": nodeid,
                    "duration": 0.0,
                    "outcome": [],
                    "runlog": {},
                    "longrepr": "",
                }
            if isinstance(result["longrepr"], list):
                results["tests"][nodeid]["longrepr"] += "\n".join(
                    [str(item) for item in result["longrepr"]]
                )
            elif isinstance(result["longrepr"], dict):
                results["tests"][nodeid]["longrepr"] += longrepr_to_string(
                    result["longrepr"]
                )

            results["tests"][nodeid]["duration"] += result["duration"]
            results["tests"][nodeid]["outcome"] += [result["outcome"]]

            runlog = runlog_from_path(path / sanitize_ident(result["nodeid"]))
            if runlog:
                results["tests"][nodeid]["runlog"] = runlog

    for nodeid, testcase in results["tests"].items():
        results["status"]["total"] += 1
        for key in ["failed", "skipped", "passed"]:
            if key in testcase["outcome"]:
                results["status"][key] += 1
                break

    if results["status"]["total"]:
        return results

    return {}


def artifacts_in_path(path: Path):
    """Returns a list of paths to artifacts"""

    if not path.exists():
        return []

    artifacts = []
    for artifact_dir in Path(path).rglob("artifacts"):
        for artifact in artifact_dir.rglob("*"):
            artifacts.append(artifact.relative_to(path))

    return sorted(artifacts)


def process_workflow_output(args, cijoe):
    workflow_state = dict_from_yamlfile(args.output / "workflow.state")
    workflow_state["config"] = cijoe.config.options
    workflow_state["artifacts"] = artifacts_in_path(args.output)

    #    workflow_state["artifacts"] = artifacts_in_path(
    #        args.output, args.output / "artifacts"
    #    )

    for step in workflow_state["steps"]:
        if "extras" not in step:
            step["extras"] = {}

        step_path = args.output / step["id"]
        if not step_path.exists():
            continue

        # artifacts = artifacts_in_path(args.output, step_path / "artifacts")
        # artifacts = artifacts_in_path(step_path / "artifacts")
        # if artifacts:
        #    step["extras"]["artifacts"] = artifacts

        runlog = runlog_from_path(step_path)
        if runlog:
            step["extras"]["runlog"] = runlog

        testreport = testreport_from_file(step_path)
        if testreport:
            step["extras"]["testreport"] = testreport

    return workflow_state



import pytest

import cijoe.linux.kmemleak as kmemleak


def skip_disabled(cijoe):
    """Skip testing when configuration is missing or module not enabled"""

    kmemleak = cijoe.config.options.get("kmemleak", None)
    if not kmemleak:
        pytest.skip(reason="missing: config['kmemleak'] configuration")

    enabled = kmemleak.get("enabled", False)
    if not enabled:
        pytest.skip(reason="kmemleak is disabled: !config['kmemleak']['enabled']")


def test_clear(cijoe):
    skip_disabled(cijoe)

    err, state = kmemleak.clear(cijoe)
    assert not err, "Failed clearing kmemleak"


def test_scan(cijoe):
    skip_disabled(cijoe)

    err, state = kmemleak.scan(cijoe)
    assert not err, "Failed scanning kmemleak"


def test_cat(cijoe):
    skip_disabled(cijoe)

    err, state = kmemleak.cat(cijoe)
    assert not err, "Failed cat'ing kmemleak"



"""
    Wraps qemu binaries: system, qemu-img and provides "guest-control"

    NOTE: this wrapper is "local-only". That is, changing transport does not retarget
    the functionality provided here. Most of the code is utilizing Python modules such
    as shutil, pathlib, psutil, download/requests. To make this re-targetable these
    things must be done via command-line utilities. It is certainly doable, however,
    currently not a priority as the intent is to utilize qemu to produce a virtual
    machine to serve as a 'target' for tests.
"""
import logging as log
import os
import shutil
import time
from pathlib import Path
from pprint import pformat

import psutil
from cijoe.core.misc import download


def qemu_img(cijoe, args=""):
    """Helper function wrapping around 'qemu-img'"""

    return cijoe.run_local(f"{cijoe.config.options['qemu']['img_bin']} {args}")


def qemu_system(cijoe, args=""):
    """Wrapping the qemu system binary"""

    return cijoe.run_local(f"{cijoe.config.options['qemu']['system_bin']} {args}")


class Guest(object):
    def __init__(self, cijoe, config, guest_name=None):
        """."""

        qemu_config = config.options.get("qemu", {})
        qemu_guests = qemu_config.get("guests", [])
        if not (qemu_config and qemu_guests):
            raise ValueError(f"Invalid qemu_config({pformat(qemu_config)})")

        guest_name = qemu_config.get("default_guest", list(qemu_guests.keys())[0])
        if not guest_name:
            raise ValueError(f"Invalid qemu_config({pformat(qemu_config)})")

        guest_config = qemu_guests.get(guest_name, None)
        if not guest_config:
            raise ValueError(f"Invalid qemu_config({pformat(qemu_config)})")

        guest_path = guest_config.get("path", None)
        if not (guest_config and guest_path):
            log.error(f"invalid qemu_config({pformat(qemu_config)})")
            raise ValueError("Invalid configuration")

        self.cijoe = cijoe
        self.qemu_config = qemu_config
        self.guest_config = guest_config
        self.guest_path = Path(guest_path).resolve()
        self.boot_img = self.guest_path / "boot.img"
        self.seed_img = self.guest_path / "seed.img"
        self.pid = self.guest_path / "guest.pid"
        self.monitor = self.guest_path / "monitor.sock"
        self.serial = self.guest_path / "serial.output"

    def image_create(self, filename, fmt="raw", size="8GB"):
        """
        Creates an image-file in the guest_path. Returns 0 on succes, errno to
        indicate the error.
        """

        img_path = self.guest_path / filename
        err, _ = qemu_img(self.cijoe, f"create -f {fmt} {img_path} {size}")

        return err

    def is_initialized(self):
        """Check that the guest is initialized"""

        return self.guest_path.exists()

    def is_running(self):
        """Check whether the guest is running"""

        pid = self.get_pid()

        return pid and psutil.pid_exists(pid)

    def get_pid(self):
        """Returns pid from 'guest.pid', returns 0 when 'guest.pid' is not found"""

        if not self.pid.exists():
            return 0

        with self.pid.open() as pidfile:
            pid = int(pidfile.read().strip())

        return pid

    def initialize(self):
        """Create a 'home' for the guest'"""

        os.makedirs(self.guest_path, exist_ok=True)

    def is_up(self, timeout=120):
        """Wait at most 'timeout' seconds for the guest to print 'login' to serial"""

        if not self.is_running():
            return False

        began = time.time()
        while True:
            enter = time.time()
            try:
                with self.serial.open() as serialfile:
                    if "login:" in serialfile.read():
                        time.sleep(10)
                        return True
            except Exception as exc:
                log.error(f"{exc}")

            now = time.time()
            elapsed_iter = now - enter
            elapsed_total = now - began

            if elapsed_iter < 5.0:
                time.sleep(5.0 - elapsed_iter)
            if elapsed_total > timeout:
                return False

    def start(self, daemonize=True, extra_args=[]):
        """."""

        args = []

        system_args = self.guest_config.get("system_args", {})

        # Key-word arguments: config.qemu.guest.system_args.kwa
        for key, value in system_args.get("kwa", {}).items():
            args.append(f"-{key}")
            if isinstance(value, dict):
                args.append(next((f"{opt}={val}" for opt, val in value.items())))
            else:
                args.append(str(value))

        # Magic: when 'boot.img' exists, add it is as boot-drive
        if self.boot_img.exists():
            args += [
                "-blockdev",
                f"qcow2,node-name=boot,file.driver=file,file.filename={self.boot_img}",
            ]
            args += ["-device", "virtio-blk-pci,drive=boot"]

        # Process Management stuff
        args += ["-pidfile", str(self.pid)]
        args += ["-monitor", f"unix:{self.monitor},server,nowait"]

        if daemonize:
            args += ["-display", "none"]
            args += ["-serial", f"file:{self.serial}"]
            args += ["-daemonize"]
        else:
            args += ["-nographic"]
            args += ["-serial", "mon:stdio"]

        #
        # Managed: config.qemu.guest.system_args.host_share
        #
        host_share = system_args.get("host_share", None)
        if host_share:
            host_share = Path(host_share).resolve()
            args += [
                "-virtfs",
                "fsdriver=local,id=fsdev0,security_model=mapped,mount_tag=hostshare"
                f",path={host_share}",
            ]

        #
        # Managed: config.qemu.guest.system_args.tcp_forward
        #
        ports = system_args.get("tcp_forward", None)
        if ports:
            args += [
                "-netdev",
                f"user,id=n1,ipv6=off,hostfwd=tcp::{ports['host']}-:{ports['guest']}",
            ]
            args += ["-device", "virtio-net-pci,netdev=n1"]

        #
        # Arguments provided via wrapper-class
        #
        args += extra_args

        #
        # Raw: config.qemu.guest.system_args.raw
        #
        args += [system_args.get("raw", "")]

        err, _ = qemu_system(self.cijoe, " ".join(args))

        return err

    def kill(self):
        """Shutdown qemu guests by killing the process using the 'guest.pid'"""

        err = 0

        try:
            pid = self.get_pid()
            if pid:
                qemu_proc = psutil.Process(pid)
                qemu_proc.terminate()

                gone, alive = psutil.wait_procs([qemu_proc], timeout=3)
                for proc in alive:
                    proc.kill()
        except psutil.NoSuchProcess:
            log.info("Got 'NoSuchProcess', that is OK, continue.")

        return err

    def init_using_cloudinit(self):
        """Provision a guest OS using cloudinit"""

        self.kill()  # Ensure the guest is *not* running
        self.initialize()  # Ensure the guest has a "home"

        # Ensure the guest has a cloudinit-image available for "installation"
        cloudinit = self.guest_config.get("init_using_cloudinit", {})
        if not cloudinit:
            log.error("missing config([qemu.guest.init_using_cloudinit])")
            return 1

        img = cloudinit.get("img", None)
        if not img:
            log.error("missing config([qemu.guest.init_using_cloudinit.img])")
            return 1

        img = Path(img).resolve()
        if not img.exists():
            url = cloudinit.get("url", None)
            if not url:
                log.error("missing config([qemu.guest.init_using_cloudinit.url])")
                return 1

            img.parent.mkdir(parents=True, exist_ok=True)
            err, path = download(url, img)
            if err:
                log.error(f"download({url}), {path}: failed")
                return err

        # Create the boot.img based on cloudinit_img
        shutil.copyfile(str(img), str(self.boot_img))
        qemu_img(self.cijoe, f"resize {self.boot_img} 10G")

        # Create seed.img, with data and meta embedded
        metadata_path = shutil.copyfile(
            cloudinit["meta"], self.guest_path / "meta-data"
        )
        userdata_path = shutil.copyfile(
            cloudinit["user"], self.guest_path / "user-data"
        )

        # Inject the "pubkey" from config
        if "pubkey" in cloudinit:
            with Path(cloudinit["pubkey"]).resolve().open() as kfile:
                pubkey = kfile.read()
            with userdata_path.open("a") as userdatafile:
                userdatafile.write("ssh_authorized_keys:\n")
                userdatafile.write(f"- {pubkey}\n")

        # This uses mkisofs instead of cloud-localds, such that it works on
        # macOS and Linux, the 'mkisofs' should be available with 'cdrtools'
        cloud_cmd = " ".join(
            [
                "mkisofs",
                "-output",
                str(self.seed_img),
                "-volid",
                "cidata",
                "-joliet",
                "-rock",
                str(userdata_path),
                str(metadata_path),
            ]
        )
        err, _ = self.cijoe.run_local(cloud_cmd)

        # Additional args to pass to the guest when starting it
        system_args = []
        system_args += ["-drive", f"file={self.seed_img},if=virtio,format=raw"]

        err = self.start(daemonize=False, extra_args=system_args)
        if err:
            log.error("failed starting...")
            return err

        return 0

    def init_using_bootimage(self):
        """Provision a guest OS using a bootable disk-image"""

        self.kill()  # Ensure the guest is *not* running
        self.initialize()  # Ensure the guest has a "home"

        # Ensure the guest has an image available to boot from
        boot = self.guest_config.get("init_using_bootimage", {})
        boot["img"] = Path(boot["img"]).resolve()

        if not boot["img"].exists():
            os.makedirs(boot["img"].parent, exist_ok=True)
            err, path = download(boot["url"], str(boot["img"]))
            if err:
                log.error(f"download({boot['url']}), {path}: failed")
                return err

        # Create the boot.img based on cloudinit_img
        shutil.copyfile(str(boot["img"]), str(self.boot_img))
        qemu_img(self.cijoe, f"resize {self.boot_img} 10G")

        return 0



#!/usr/bin/env python3
"""
Kill a qemu guest
=================

Retargetable: False
-------------------
"""
from cijoe.qemu.wrapper import Guest


def main(args, cijoe, step):
    """Kill a qemu guest"""

    guest = Guest(cijoe, cijoe.config)

    return guest.kill()



import pytest


@pytest.fixture
def nvme(cijoe, capsys):
    nvme = cijoe.get_config(subject="nvme")
    if not nvme:
        return None

    for device in nvme.get("devices", []):
        return device


@pytest.fixture
def xnvme(cijoe, capsys):
    nvme = cijoe.get_config(subject="nvme")
    if not nvme:
        return None

    for device in nvme.get("devices", []):
        return device
